# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright Jon Hill, University of York, jon.hill@york.ac.uk

"""
hrds contains the following components:

    * buffer - functions to create buffer files
    * hdrs - the main hdrs object.
    * RasterInterpolator - objects to interpolate individual rasters

"""

__version__ = "0.1.3"

from .raster import RasterInterpolator   # NOQA
from .raster_buffer import CreateBuffer  # NOQA
from .hrds import HRDS                   # NOQA
